import json

import ruamel.yaml
from identify import identify

from .errors import BadFileTypeError

yaml = ruamel.yaml.YAML(typ="safe")


class InstanceLoader:
    def __init__(self, filenames, default_filetype=None):
        self._filenames = filenames
        self._default_ft = default_filetype

    @property
    def _default_loadfunc(self):
        if not self._default_ft:
            return None
        if self._default_ft.lower() == "json":
            return json.load
        return yaml.load

    def iter_files(self):
        for fn in self._filenames:
            tags = identify.tags_from_path(fn)
            if "yaml" in tags:
                loadfunc = yaml.load
            elif "json" in tags:
                loadfunc = json.load
            else:
                loadfunc = self._default_loadfunc

            # TODO: handle this by storing it in the errors map
            if not loadfunc:
                raise BadFileTypeError(
                    f"cannot check {fn} as it is neither yaml nor json"
                )

            with open(fn) as fp:
                yield (fn, loadfunc(fp))


def instance_loader_from_args(args):
    return InstanceLoader(args.instancefiles, default_filetype=args.default_filetype)



import contextlib
import os
import platform
import shutil
import tempfile
import time
import typing as t
import urllib.request


class CacheDownloader:
    _LASTMOD_DEFAULT = "Sun, 01 Jan 1970 00:00:01 GMT"
    _LASTMOD_FMT = "%a, %d %b %Y %H:%M:%S %Z"

    # changed in v0.5.0
    # original cache dir was "jsonschema_validate"
    # this will let us do any other caching we might need in the future in the same
    # cache dir (adjacent to "downloads")
    _CACHEDIR_NAME = os.path.join("check_jsonschema", "downloads")

    def __init__(
        self,
        file_url: str,
        filename: t.Optional[str] = None,
        cache_dir: t.Optional[str] = None,
        disable_cache: bool = False,
    ):
        self._file_url = file_url
        self._filename = filename or file_url.split("/")[-1]
        self._cache_dir = cache_dir or self._compute_default_cache_dir()
        self._disable_cache = disable_cache

    def _compute_default_cache_dir(self) -> t.Optional[str]:
        sysname = platform.system()

        # on windows, try to get the appdata env var
        # this *could* result in cache_dir=None, which is fine, just skip caching in
        # that case
        if sysname == "Windows":
            cache_dir = os.getenv("LOCALAPPDATA", os.getenv("APPDATA"))
        # macOS -> app support dir
        elif sysname == "Darwin":
            cache_dir = os.path.expanduser("~/Library/Caches")
        # default for unknown platforms, namely linux behavior
        # use XDG env var and default to ~/.cache/
        else:
            cache_dir = os.getenv("XDG_CACHE_HOME", os.path.expanduser("~/.cache"))

        if cache_dir:
            cache_dir = os.path.join(cache_dir, self._CACHEDIR_NAME)

        return cache_dir

    def _lastmod_from_conn(self, conn) -> float:
        return time.mktime(
            time.strptime(
                conn.headers.get("last-modified", self._LASTMOD_DEFAULT),
                self._LASTMOD_FMT,
            )
        )

    def _cache_hit(self, cachefile, conn):
        # no file? miss
        if not os.path.exists(cachefile):
            return False

        # compare mtime on any cached file against the remote last-modified time
        # it is considered a hit if the local file is at least as new as the remote file
        local_mtime = os.path.getmtime(cachefile)
        remote_mtime = self._lastmod_from_conn(conn)
        return local_mtime >= remote_mtime

    @contextlib.contextmanager
    def _urlopen(self):
        with urllib.request.urlopen(self._file_url) as conn:
            yield conn

    def _write(self, conn, dest):
        # download to a temp file and then move to the dest
        # this makes the download safe if run in parallel (parallel runs
        # won't create a new empty file for writing and cause failures)
        fp = tempfile.NamedTemporaryFile(mode="wb", delete=False)
        fp.write(conn.read())
        fp.close()
        shutil.copy(fp.name, dest)
        os.remove(fp.name)

    def _download(self):
        os.makedirs(self._cache_dir, exist_ok=True)
        dest = os.path.join(self._cache_dir, self._filename)

        with self._urlopen() as conn:
            # check to see if we have a file which matches the connection
            # only download if we do not (cache miss, vs hit)
            if not self._cache_hit(dest, conn):
                self._write(conn, dest)

        return dest

    @contextlib.contextmanager
    def open(self):
        if not self._cache_dir or self._disable_cache:
            with urllib.request.urlopen(self._file_url) as fp:
                yield fp
        else:
            cached_file = self._download()
            with open(cached_file, "r") as fp:
                yield fp



import argparse


# support passing through an argparser class to support tests
def get_parser(cls=None):
    parser = cls() if cls is not None else argparse.ArgumentParser()
    parser.add_argument(
        "--schemafile",
        help=(
            "The path to a file containing the jsonschema to use or an "
            "HTTP(S) URI for the schema. If a remote file is used, "
            "it will be downloaded and cached locally based on mtime. "
            "This option is required unless --builtin-schema is passed."
        ),
    )
    parser.add_argument(
        "--builtin-schema",
        help=(
            "The name of an internal schema to use for `--schemafile`. "
            "Schema names can be found in project documentation."
        ),
    )
    parser.add_argument(
        "--failover-builtin-schema",
        help=(
            "Failover to a specific builtin schema if the schemafile is remote and "
            "cannot be fetched."
        ),
        type=str.lower,
    )
    parser.add_argument(
        "--no-cache",
        action="store_true",
        default=False,
        help="Disable schema caching. Always download remote schemas.",
    )
    parser.add_argument(
        "--disable-format",
        action="store_true",
        default=False,
        help="Disable all format checks in the schema. "
        "Because JSON Schema does not require that the format property do any "
        "validation, files which validate with other tools may fail to validate "
        "without this flag.",
    )
    parser.add_argument(
        "--cache-filename",
        help=(
            "The name to use for caching a remote schema. "
            "Defaults to the last slash-delimited part of the URI."
        ),
    )
    parser.add_argument(
        "--default-filetype",
        help=(
            "A default filetype to assume when a file is not detected as JSON or YAML"
        ),
        type=str.lower,
        choices=("json", "yaml"),
    )
    parser.add_argument(
        "--traceback-mode",
        help=(
            "Set the mode of presentation for error traces. "
            "Defaults to shortened tracebacks."
        ),
        type=str.lower,
        default="short",
        choices=("full", "short"),
    )
    parser.add_argument("instancefiles", nargs="+", help="JSON or YAML files to check.")
    return parser


def parse_args(args=None, cls=None):
    parser = get_parser(cls=cls)
    args = parser.parse_args(args)
    if args.schemafile and args.builtin_schema:
        parser.error("--schemafile and --builtin-schema are mutually exclusive")
    if not args.schemafile:
        if not args.builtin_schema:
            parser.error("Either --schemafile or --builtin-schema must be provided")
        if args.failover_builtin_schema:
            parser.error(
                "--failover-builtin-schema is incompatible with --builtin-schema"
            )
    return args



from .errors import BadFileTypeError, SchemaParseError, UnsupportedUrlScheme
from .instance import InstanceLoader, instance_loader_from_args
from .schema import BuiltinSchemaLoader, SchemaLoader, schema_loader_from_args

__all__ = (
    "BadFileTypeError",
    "SchemaParseError",
    "UnsupportedUrlScheme",
    "BuiltinSchemaLoader",
    "SchemaLoader",
    "InstanceLoader",
    "schema_loader_from_args",
    "instance_loader_from_args",
)
